/*****************************************************************************
/*
                                  WASD.h

Main header file for WASD VMS HTTP daemon.


VERSION HISTORY
---------------
19-SEP-98  MGD  v5.3.0, disperse to module headers :^)
14-APR-98  MGD  v5.0.1, XSSI extensions
07-JAN-98  MGD  v5.0.0, HTTPd version 5 (Secure Sockets Layer, using SSLeay)
05-OCT-97  MGD  v4.5.0, file cache
06-SEP-97  MGD  v4.4.0, multiple server host names and ports
08-JUN-97  MGD  v4.2.0, CGIplus
01-FEB-97  MGD  v4.0.0, version 4 :^)
01-AUG-96  MGD  v3.3.0, quite a few changes since 3.0.n
01-DEC-95  MGD  v3.0.0, HTTPd version 3
20-DEC-94  MGD  v2.0.0, multi-threaded version
20-JUN-94  MGD  v1.0.0, single-threaded version
*/
/*****************************************************************************/

#ifndef WASD_H_LOADED
#define WASD_H_LOADED 1

/* IP_NONE  is defined when compiling MapUrl.c for stand-alone use */
#ifndef IP_NONE
/* if not NETLIB then UCX! */
#ifndef IP_NETLIB
#  undef IP_UCX
#  define IP_UCX 1
#endif
#endif

#include <descrip.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>
#include <rmsdef.h>

/* Internet-related header files */
#ifdef IP_UCX
#   include <socket.h>
#   include <in.h>
#   include <netdb.h>
#   include <inet.h>
#   ifdef IP_TCPWARE
#      include "tcpware_include:ucx$inetdef.h"
#   else
#      include <ucx$inetdef.h>
#   endif
#endif

#ifdef IP_NETLIB
#   include "netlibdef.h"
#endif

#ifdef __ALPHA
#   pragma nomember_alignment
#endif

/**********/
/* macros */
/**********/

#define boolean int
#define true 1
#define false 0

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))

#define FI_LI WASD_MODULE, __LINE__

#define AuthConfigFileName "HTTPD$AUTH"
#define ConfigFileName "HTTPD$CONFIG"
#define MapFileName "HTTPD$MAP"
#define MsgFileName "HTTPD$MSG"

/* this value must also include connections in a "keep-alive" state */
#define DefaultMaxConcurrentConnections 20

#define DefaultInputTimeoutMinutes 1
#define DefaultOutputTimeoutMinutes 10
#define DefaultKeepAliveTimeoutSeconds 0

#define DefaultDclSysOutputSize 4096
#define DefaultFileBufferSize 1024
#define DefaultNetReadBufferSize 2048
#define DefaultOutputBufferSize 4096

#define HttpdInternalScriptEcho "/echo"
#define HttpdInternalScriptTree "/tree"
#define HttpdInternalScriptUpd "/upd"
#define HttpdInternalScriptWhere "/where"
#define HttpdInternalScriptXray "/xray"

#define HttpdInternalPasswordChange "/httpd/-/change/"

#define HttpdInternal "/httpd/-/"

#define HttpdInternalAdmin "/httpd/-/admin/"

#define HttpdInternalControl "/httpd/-/admin/control/"
#define HttpdInternalControlRestart "/httpd/-/admin/control/restart"
#define HttpdInternalControlRestartNow "/httpd/-/admin/control/restartNOW"
#define HttpdInternalControlExit "/httpd/-/admin/control/exit"
#define HttpdInternalControlExitNow "/httpd/-/admin/control/exitNOW"
#define HttpdInternalControlLogOn "/httpd/-/admin/control/logon"
#define HttpdInternalControlLogOff "/httpd/-/admin/control/logoff"
#define HttpdInternalControlLogFlush "/httpd/-/admin/control/logflush"
#define HttpdInternalControlCacheOn "/httpd/-/admin/control/cacheon"
#define HttpdInternalControlCacheOff "/httpd/-/admin/control/cacheoff"
#define HttpdInternalControlCachePurge "/httpd/-/admin/control/cachepurge"

#define HttpdInternalGraphic "/httpd/-/admin/graphic/"
#define HttpdInternalGraphicActivity "/httpd/-/admin/graphic/activity"

#define HttpdInternalReport "/httpd/-/admin/report/"
#define HttpdInternalReportActivity "/httpd/-/admin/report/activity"
#define HttpdInternalReportCache "/httpd/-/admin/report/cache"
#define HttpdInternalReportPaths "/httpd/-/admin/report/paths"
#define HttpdInternalReportRules "/httpd/-/admin/report/rules"
#define HttpdInternalReportSSL "/httpd/-/admin/report/SSL"
#define HttpdInternalReportWatch "/httpd/-/admin/report/watch"

#define HttpdInternalRevise "/httpd/-/admin/revise/"
#define HttpdInternalReviseAuth "/httpd/-/admin/revise/auth/"
#define HttpdInternalReviseConfig "/httpd/-/admin/revise/config"
#define HttpdInternalReviseMessages "/httpd/-/admin/revise/messages"
#define HttpdInternalRevisePaths "/httpd/-/admin/revise/paths"
#define HttpdInternalReviseRules "/httpd/-/admin/revise/rules"

/* helps ensure a client doesn't get into an infinite loop */
#define MaxKeepAliveCount 100

/*
   Attempt to reduce dynamic memory fragmentation within C RTL by
   maintaining a minimum size chunk to be allocated.  Hopefully this
   will prevent lots of small chunks having to be garbage-collected
   or otherwise processed (as has sometimes been observed ... at least
   this is the best explanation I can come up with).
*/
#define HEAP_MIN_CHUNK_SIZE 256

/* control pre-expired responses */
#define PRE_EXPIRE_ADMIN            1
#define PRE_EXPIRE_DELETE_ON_CLOSE  1
#define PRE_EXPIRE_INDEX_OF         0
#define PRE_EXPIRE_MAPPING_RULES    0
#define PRE_EXPIRE_MENU             1
#define PRE_EXPIRE_PUT              1
#define PRE_EXPIRE_SSI              0
#define PRE_EXPIRE_UPD              1
#define PRE_EXPIRE_UPD_TREE_OF      1
#define PRE_EXPIRE_WATCH            0

/* constants used to identify HTTP methods */
#define HTTP_METHOD_DELETE  0x00000001
#define HTTP_METHOD_GET     0x00000002
#define HTTP_METHOD_HEAD    0x00000004
#define HTTP_METHOD_POST    0x00000008
#define HTTP_METHOD_PUT     0x00000010

/* constants for controlling logging */
#define LOGGING_BEGIN  1
#define LOGGING_END    2
#define LOGGING_OPEN   3
#define LOGGING_CLOSE  4
#define LOGGING_FLUSH  5
#define LOGGING_ENTRY  6

/* constants used by HTTPd supervisor timer functionality */
#define TIMER_INPUT      1
#define TIMER_OUTPUT     2
#define TIMER_KEEPALIVE  3

/* constants used by the description module */
#define DESCRIPTION_ALL         0x01
#define DESCRIPTION_TEXT_PLAIN  0x02
#define DESCRIPTION_TEXT_HTML   0x04
/* this is returned in the first character the file type contains none */
#define DESCRIPTION_IMPOSSIBLE  0xff

/* default directory listing layout */
#define DEFAULT_DIR_LAYOUT "I__L__R__S__D"

/* "http:" or "https:" service and request */
#define SCHEME_HTTP  1
#define SCHEME_HTTPS 2

/* if non-zero then imposes a maximum byte count on a request body */
#define MaxContentLength 0

/* maximum Kbytes (1024) that can be in a body to be POSTed or PUTted */
#define DefaultPutMaxKbytes 250
/* number of versions retained for files POSTed or PUTed */
#define DefaultPutVersionLimit 3

/* header fields for HTTP/1.0 keep alive response */
#define KeepAliveHttpHeader "Connection: Keep-Alive\r\nKeep-Alive:\r\n"

/* if true, stops HTTPd server account with SYSPRV PUT(etc)ing - DANGEROUS */
#define PUT_DISALLOW_SYSPRV 0

#define MaxConfigHomePages 16
#define MaxConfigReadMeFiles 16
#define MaxConfigScriptRunTimes 16

#define SCRIPT_NAME_SIZE 128
#define METHOD_NAME_SIZE 16

/* W:RE,G:RE,O:RWED,S:RWED */
#define PUT_DEFAULT_FILE_PROTECTION 0xaa00

#define REQUEST_REDIRECTION_MAX 5

/*************/
/* durations */
/*************/

/* 1000 = 1mS, 100 = 100uS, 10 = 10Us, 1 = 1uS (for really fast systems :^) */
#define DURATION_UNITS_USECS 1000

#if DURATION_UNITS_USECS == 1
#  define DURATION_DECIMAL_PLACES 6
#elif DURATION_UNITS_USECS == 10
#  define DURATION_DECIMAL_PLACES 5
#elif DURATION_UNITS_USECS == 100
#  define DURATION_DECIMAL_PLACES 4
#else
   /* defaults to milliseconds */
#  undef DURATION_UNITS_USECS
#  define DURATION_UNITS_USECS 1000
#  define DURATION_DECIMAL_PLACES 3
#endif

#define USECS_IN_A_SECOND 1000000

/*********************/
/* functional macros */
/*********************/

/* is it linear-white-space? (avoids the call overhead of an isspace()) */
#define ISLWS(c) (c == ' ' || c == '\t')

/* end of CR or LF terminated line? */
#define EOL(c) (c == '\r' || c == '\n' || !*cptr)
#define NOTEOL(c) (c != '\r' && c != '\n' && *cptr)

#define STRCPY(src,dst,errfunc) \
   cptr = src; \
   zptr = (sptr = dst) + sizeof(src); \
   while (*cptr && sptr < zptr) *sptr++ = *cptr++; \
   if (sptr >= zptr) \
   { \
      ErrorGeneralOverflow (rqptr, FI_LI); \
      errfunc (rqptr); \
      return; \
   } \
   *sptr = '\0';

#define STRCPYLEN(src,dst,dstlen,errfunc) \
   cptr = src; \
   zptr = (sptr = dst) + sizeof(src); \
   while (*cptr && sptr < zptr) *sptr++ = *cptr++; \
   if (sptr >= zptr) \
   { \
      ErrorGeneralOverflow (rqptr, FI_LI); \
      errfunc (rqptr); \
      return; \
   } \
   *sptr = '\0'; \
   dstlen = sptr - dst;

/*************************************************/
/* macros for development timing in microseconds */
/*************************************************/

/* 0..9 timers available */

#define USEC_TIMER_STORAGE \
   unsigned long  uSecSubx2 = 2, \
                  uSecEdiv10 = 10; \
   unsigned long  uSecResult, \
                  uSecDuration, \
                  uSecRemainder; \
   unsigned long  uSecBeginQuad [10][2], \
                  uSecEndQuad [10][2];

#define USEC_TIMER_STORAGE_EXTERN \
   extern unsigned long  uSecSubx2, \
                         uSecEdiv10; \
   extern unsigned long  uSecResult, \
                         uSecDuration, \
                         uSecRemainder; \
   extern unsigned long  uSecBeginQuad [10][2], \
                         uSecEndQuad [10][2];

#define USEC_TIMER_BEGIN(IDX) \
{ \
   sys$gettim (&uSecBeginQuad[IDX]); \
}

#define USEC_TIMER_END(IDX,WHERE) \
{ \
   sys$gettim (&uSecEndQuad[IDX]); \
   lib$subx (&uSecEndQuad[IDX], &uSecBeginQuad[IDX], &uSecResult, &uSecSubx2); \
   lib$ediv (&uSecEdiv10, &uSecResult, &uSecDuration, &uSecRemainder); \
   fprintf (stdout, "%s uS: %d\n", WHERE, uSecDuration); \
}

/*******************/
/* data structures */
/*******************/

# pragma member_alignment __save
# pragma nomember_alignment

struct VmsItem
{
   unsigned short  buf_len;
   unsigned short  item;
   unsigned char   *buf_addr;
   unsigned short  *ret_len;
};

struct AnExitHandler
{
   unsigned long  Flink;
   unsigned long  HandlerAddress;
   unsigned long  ArgCount;
   unsigned long  ExitStatusPtr;
};

struct AnIOsb
{
   unsigned short  Status;
   unsigned short  Count;
   unsigned long  Unused;
};

struct MbxSenseIOsb
{
   unsigned short  Status;
   unsigned short  MessageCount;
   unsigned long  MessageBytes;
};

#pragma member_alignment __restore

/*
   Store these structures naturally-aligned on AXP.
   Uses a bit more storage but makes access as efficient as possible.
*/

#ifdef __ALPHA
#   pragma member_alignment __save
#   pragma member_alignment
#endif

/* per-request memory heap */
struct HeapStruct
{
   struct HeapStruct  *PrevPtr;
   struct HeapStruct  *NextPtr;
   unsigned int  ChunkSize;
   /* the actual data follows immediately after the length */
};

/*
This structure is used to both store and to point at the various components
depending on the context.
*/

struct ContentTypeStruct
{
   struct ContentTypeStruct  *NextPtr,
                             *HashCollisionPtr;
   boolean  ContentTypeUnknown;
   char  *SuffixPtr;
   char  *ContentTypePtr;
   char  *AutoScriptNamePtr;
   char  *ContentDescriptionPtr;
   char  *IconPtr;
   /* string data is stored here */
};

/*
This structure is used to both store and to point at the various components
depending on the context.
*/

struct IconStruct
{
   struct IconStruct  *NextPtr,
                      *HashCollisionPtr;
   /* string data is stored here */
};

struct  ListEntryStruct
{
   struct ListEntryStruct  *PrevPtr;
   struct ListEntryStruct  *NextPtr;
};

struct  ListHeadStruct
{
   struct ListEntryStruct  *HeadPtr;
   struct ListEntryStruct  *TailPtr;
   int  EntryCount;
};

/* these macros operate on the list head */
#define LIST_HEAD(lp) ((char*)((struct ListHeadStruct*)lp)->HeadPtr)
#define LIST_TAIL(lp) ((char*)((struct ListHeadStruct*)lp)->TailPtr)
#define LIST_COUNT(lp) (((struct ListHeadStruct*)lp)->EntryCount)
#define LIST_EMPTY(lp) (!((struct ListHeadStruct*)lp)->EntryCount)

/* these macros operate on list entries */
#define LIST_DATA(lp) ((char*)((char*)lp+sizeof(struct ListEntryStruct)))
#define LIST_PREV(lp) (((struct ListEntryStruct*)lp)->PrevPtr)
#define LIST_NEXT(lp) (((struct ListEntryStruct*)lp)->NextPtr)
#define LIST_HAS_PREV(lp) ((((struct ListEntryStruct*)lp)->PrevPtr) != NULL)
#define LIST_HAS_NEXT(lp) ((((struct ListEntryStruct*)lp)->NextPtr) != NULL)

/**********************************/
/* request history data structure */
/**********************************/

/* 256 bytes (a somewhat arbitrary quantity) */
struct  RequestHistoryStruct
{
   struct  ListEntryStruct  HistoryListEntry;
   unsigned long  BinaryTime [2];
   unsigned long  BytesRx,
                  BytesTx,
                  ResponseDuration;
   int  RequestScheme,
        ResponseStatusCode;
   char  *RequestPtr;
   boolean  Truncated;
   char  ClientAndRequest [256-12-8-4-4-4-4-4-4];
};

/************************/
/* module include files */
/************************/

#include "admin.h"
#include "auth.h"
#include "basic.h"
#include "cache.h"
#include "cgi.h"
#include "config.h"
#include "control.h"
#include "descr.h"
#include "dcl.h"
#include "dir.h"
#include "decnet.h"
#include "error.h"
#include "file.h"
#include "graph.h"
#include "htadmin.h"
#include "httpd.h"
#include "ismap.h"
#include "mapurl.h"
#include "menu.h"
#include "msg.h"
#include "net.h"
#include "put.h"
#include "request.h"
#include "ssi.h"
#include "stmlf.h"
#include "support.h"
#include "upd.h"
#include "vm.h"

/***************************/
/* configuration structure */
/***************************/

/* must be powers of two! */
#define  CONTENT_TYPE_HASH_TABLE_SIZE  128
#define  CONTENT_TYPE_HASH_TABLE_MASK  CONTENT_TYPE_HASH_TABLE_SIZE-1
#define  ICON_HASH_TABLE_SIZE  64
#define  ICON_HASH_TABLE_MASK  ICON_HASH_TABLE_SIZE-1

struct ConfigStruct
{
   boolean  AuthBasicEnabled,
            AuthDigestEnabled,
            CacheEnabled,
            DclFullRequest,
            DclSpawnAuthPriv,
            DirAccess,
            DirAccessSelective,
            DirNoImpliedWildcard,
            DirNoPrivIgnore,
            DirOwnerEnabled,
            DirPreExpired,
            DirReadMeBottom,
            DirReadMeTop,
            DirWildcardEnabled,
            DnsLookup,
            ErrorRecommend,
            ErrorSourceInfo,
            IncludeCommentedInfo,
            LoggingEnabled,
            LogPerService,
            MonitorEnabled,
            SsiEnabled,
            SsiAccessesEnabled,
            SsiExecEnabled;

   int  AcceptHostsLength,
        ActivityNumberOfDays,
        AuthCacheMinutes,
        AuthDigestNonceGetLifeTime,
        AuthDigestNoncePutLifeTime,
        AuthFailureLimit,
        AuthRevalidateUserMinutes,
        Busy,
        CacheEntriesMax,
        CacheChunkKBytes,
        CacheFileKBytesMax,
        CacheFrequentHits,
        CacheFrequentSeconds,
        CacheHashTableEntries,
        CacheTotalKBytesMax,
        CacheValidateSeconds,
        DclCgiPlusLifeTime,
        DclSubprocessHardLimit,
        DclSubprocessSoftLimit,
        DclZombieLifeTime,
        DECnetConnectListMax,
        DECnetReuseLifeTime,
        DirDescriptionLines,
        ErrorReportPathLength,
        HomePageCount,
        InputTimeoutMinutes,
        KeepAliveTimeoutSeconds,
        KeywordSearchScriptLength,
        LogExcludeHostsLength,
        OutputTimeoutMinutes,
        PutMaxKbytes,
        PutVersionLimit,
        ReadMeFileCount,
        RejectHostsLength,
        RequestHistory,
        ScriptRunTimeCount,
        ServerPort,
        ServiceLength,
        StreamLfConversionMaxKbytes;

   int  ScriptRunTimeLengthArray [MaxConfigScriptRunTimes];

   struct ContentTypeStruct  
      *ContentTypeHashTable [CONTENT_TYPE_HASH_TABLE_SIZE];
   struct IconStruct *IconHashTable [ICON_HASH_TABLE_SIZE];

   char  CharsetDefault [32],
         DefaultDirLayout [32],
         ErrorReportPath [128],
         KeywordSearchScript [128],
         LogFileName [128],
         LogFormat [128],
         LogNaming [32],
         LogPeriod [32];

   char  *AcceptHostsPtr,
         *DefaultPost,
         *LogExcludeHostsPtr,
         *RejectHostsPtr,
         *ServicePtr;

   char  *HomePageArray [MaxConfigHomePages],
         *ReadMeFileArray [MaxConfigReadMeFiles],
         *ScriptRunTimeArray [MaxConfigScriptRunTimes];

   struct ContentTypeStruct  *ContentTypeListHeadPtr,
                             *ContentTypeListTailPtr;
   char  *ContentTypeDefaultPtr;
   int  ContentTypeStructOverhead;

   struct IconStruct  *IconListHeadPtr,
                      *IconListTailPtr;
   int  IconStructOverhead;

   /* only used while reading configuration from file */
   int  LineNumber,
        ProblemCount,
        ProblemReportLength;
   char  *ProblemReportPtr;
   char  DirectiveName [32],
         LoadFileName [256];
   unsigned long  LoadBinTime [2],
                  RevBinTime [2];

   /* only used when reporting or changing configuration */
   void  *RequestPtr;
};

/*************************/
/* per-server accounting */
/*************************/

/* as of v4.4 this structure is now > 255 bytes requiring two logicals */

struct AccountingStruct
{
   /* this is used to check for changes in HTTPd version requiring zeroing */
   unsigned long  AccountingVersion;
   /* this MUST be the first member (or so) to allow zeroing of the rest */
   unsigned long  ZeroedCount;

   unsigned long  AuthBasicCount,
                  AuthAuthorizedCount,
                  AuthDigestCount,
                  AuthHtDatabaseCount,
                  AuthNotAuthenticatedCount,
                  AuthNotAuthorizedCount,
                  AuthVmsCount,
                  CacheHitCount,
                  CacheHitNotModifiedCount,
                  CacheLoadCount,
                  ConnectAcceptedCount,
                  ConnectCount,
                  ConnectSslCount,
                  ConnectCurrent,
                  ConnectPeak,
                  ConnectRejectedCount,
                  ConnectTooBusyCount,
                  DclCgiPlusReusedCount,
                  DclSpawnCount,
                  DoAutoScriptCount,
                  DoCgiPlusScriptCount,
                  DoDclCommandCount,
                  DoDECnetCount,
                  DoDECnetReuseCount,
                  DoDECnetCgiCount,
                  DoDECnetCgiReuseCount,
                  DoDECnetOsuCount,
                  DoDECnetOsuReuseCount,
                  DoDirectoryCount,
                  DoIsMapCount,
                  DoScriptCount,
                  DoFileCount,
                  DoFileNotModifiedCount,
                  DoMenuCount,
                  DoPutCount,
                  DoServerAdminCount,
                  DoSsiCount,
                  DoUpdateCount,
                  LastExitStatus,
                  MethodDeleteCount,
                  MethodGetCount,
                  MethodHeadCount,
                  MethodPostCount,
                  MethodPutCount,
                  MethodTraceCount,
                  RedirectLocalCount,
                  RedirectRemoteCount,
                  RequestTotalCount,
                  RequestErrorCount,
                  RequestForbiddenCount,
                  RequestKeepAliveCount,
                  RequestKeepAliveMax,
                  RequestParseCount,
                  RequestPragmaNoCacheCount,
                  RequestProxyCount,
                  ResponseDurationCount,
                  ResponseDurationMax,
                  ResponseDurationMin,
                  StartupCount,
                  StreamLfConversionCount;

   unsigned long  QuadBytesRx [2],
                  QuadBytesTx [2],
                  QuadResponseDuration [2];

   unsigned long  ResponseStatusCodeCountArray [6];
};

/*****************************/
/* per-thread data structure */
/*****************************/

struct RequestStruct
{
   /***************************/
   /* general purpose storage */
   /***************************/

   /*
      Storage areas at the beginning of this structure, specifically
      above the field 'RetainAboveZeroBelow', are retained between
      successive requests during persistent connections.
      When a "keep-alive" connection is reinitialized ready for a
      potential new request all of this structure below (and including)
      that field are zeroed, effectively allowing a new request.
   */

   /* for maintaining the list of request structures */
   struct  ListEntryStruct  RequestListEntry;

   /* pointer to service structure of host/port etc of service connected to */
   struct ServiceStruct  *ServicePtr;

   int  /* counter for input timeout */
        TimerInputCount,
        /* counter for keep-alive timeout */
        TimerKeepAliveCount,
        /* counter for output timeout */
        TimerOutputCount;

   unsigned long  /* zone ID for request heap's virtual memory management */
                  VmHeapZoneId;

#ifdef IP_UCX
   unsigned short  /* channel to client's TCP/IP device */
                   ClientChannel;
#endif

#ifdef IP_NETLIB
   unsigned long  /* socket for client's TCP/IP device */
                  ClientNetLibSocket;
#endif

   boolean  /* when the socket has been explicitly shutdown (before close) */
            ClientSocketShutdown,
            /* timers expired and the client socket has been shut/closed */
            TimerTerminated;

   struct AnIOsb  /* used, of course, for socket shutdown status */
                  NetShutdownIOsb;

   int  /* should be a short, but */
        ClientPort;

   unsigned long  /* client's binary IP address */
                  ClientIpAddress;

   unsigned char  /* internet host/domain name of client */
                  ClientHostName [128],
                  /* numeric host/domain address of client */
                  ClientIpAddressString [32];

   int  /* for life of the thread */
        AdminWatchItem,
        /* count of how many times the TCP/IP connection is reused */
        KeepAliveCount,
        /* an error is being reported from a redirect path */
        ErrorReportStatusCode,
        /* avoid redirection loops by keeping track of the number */
        RedirectionCount;

   char  /* when redirected this points the any authentication realm */
         *ErrorReportAuthRealmPtr;

   /* make it easy on ourselves, make the SSL structure pointer typeless! */
   void  *SeSoLaPtr;
   /* "http:" or "https:" (SSL) */
   int  RequestScheme;
   /* pointer to either "http:" or "https:" */
   char  *RequestSchemeNamePtr;

   /* buffers for when WATCHing network */
   void  *AdminWatchNetReadAstFuncPtr,
         *AdminWatchNetWriteAstFuncPtr;

   /* the rest of this structure can be zeroed with persistent connections */
   /**********************************/
   unsigned long  RetainAboveZeroBelow;
   /**********************************/

   boolean  /* this request is using the watch report facility */
            AdminWatchClient,
            /* authentication will be done by script */
            AuthExternal,
            /* cache is currently being loaded for this path/file */
            CacheLoading,
            /* */
            CgiAddNewline,
            /* */
            CgiBufferRecords,
            /* */
            CgiCheckedHeader,
            /* */
            CgiContentTypeText,
            /* */
            CgiScript,
            /* */
            CgiPlusScript,
            /* "temporary" file, delete it on close */
            DeleteOnClose,
            /* report any errors using the error report script */
            ErrorReportByRedirect,
            /* true if a mapping rule identifies this as a "CGIplus" script */
            IsCgiPlusScript,
            /* this request has a "Connection: Keep-Alive" token */
            KeepAliveRequest,
            /* the server has responded with "keep-alive" */
            KeepAliveResponse,
            /* this persistent connection has timed out */
            KeepAliveTimeout,
            /* escapes HTML-forbidden characters as sent to client */
            NetWriteBufferedEscapeHtml,
            /* response header has been sent to client */
            ResponseHeaderSent,
            /* response include an "Expires:" the same as "Last-Modified:" */
            ResponsePreExpired,
            /* "Pragma: no-cache" HTTP request header line encountered */
            PragmaNoCache,
            /* rule mapping, SET the path to be pre-expired */
            SetPathExpired,
            /* rule mapping, SET the path not-to-be-cached */
            SetPathNoCache,
            /* rule mapping, SET the path not-to-be-logged */
            SetPathNoLog,
            /* rule mapping, SET the path allowing stream-LF conversion */
            SetPathStmLF,
            /* an "Xray" returns full response as plain-text (diagnostic) */
            Xray;


   int  /* counter representing request activity has be incremented */
        AccountingDone,
        /* could not have come from cache, could have but didn't, did! */
        CacheInvolvement,
        /* multi-purpose, when loading and reading cache */
        CacheContentCount,
        /* length of heap memory pointed to by CgiBufferPtr */
        CgiBufferLength,
        /* remaining space in heap memory pointed to by CgiBufferPtr */
        CgiBufferRemaining,
        /* length of script end-of-file string */
        CgiEofLength,
        /* how many attempts have been made to restart a CGI script */
        CgiScriptRetryCount,
        /* count of bytes received so far in a POSTed or PUTed body */
        ContentCount,
        /* "Content-Length:" value */
        ContentLength,
        /* "Content-Type:" length */
        ContentTypeLength,
        /* size of buffer space for sys$output from DCL subprocess */
        *DclSysOutputSize,
        /* processing request header, count to find first blank line */
        HeaderConsecutiveNewLineCount,
        /* "Accept:" length */
        HttpAcceptLength,
        /* "Accept-Charset:" length */
        HttpAcceptCharsetLength,
        /* "Accept-Language:" length */
        HttpAcceptLangLength,
        /* "Forwarded:" line(s) length */
        HttpForwardedLength,
        /* length of error report text/query-string */
        ErrorReportLength,
        /* value representation of request HTTP method */
        HttpMethod,
        /* length of request header field */
        HttpAuthorizationLength,
        /* length of request header field */
        HttpRefererLength,
        /* length of request header field */
        HttpUserAgentLength,
        /* length of mapped path */
        MappedPathLength,
        /* interge representing prefered message language */
        MsgLanguage,
        /* size of network read buffer */
        NetReadBufferSize,
        /* length of path from HTTP header */
        PathInfoLength,
        /* length of query string from HTTP header */
        QueryStringLength,
        /* lenth of resource (path plus query string) from HTTp header */
        ResourceLength,
        /* number of characters up until the blank line */
        RequestHeaderLength,
        /* count of home page file names tried */
        RequestHomePageIndex,
        /* used during HEAD method to determine when header ends */
        ResponseHeaderNewlineCount,
        /* buffer for data length */
        ResponseHeaderDataLength,
        /* length of reponse header */
        ResponseHeaderLength,
        /* numeric equivalent of the "200", "302", etc., HTTP status code */
        ResponseStatusCode,
        /* VMS status of most recent task */
        TaskStatus;

   unsigned long  /* count of total bytes received from the client */
                  BytesRx,
                  /* count of total bytes sent to the client */
                  BytesTx,
                  /* buffer for hash value used when caching */
                  CacheHashValue,
                  /* response duration */
                  ResponseDuration;

   unsigned long  /* 64 bit VMS internal time, start of request processing */
                  BinaryTime [2],
                  /* 64 bit VMS time from "If-Modified-Since:" header */
                  IfModifiedSinceBinaryTime [2],
                  /* length paramater from "If-Modified-Since:" header */
                  IfModifiedSinceLength;

   unsigned short  /* component equivalent of 'BinaryTime' */
                   NumericTime [7];

   unsigned char  /* multi-purpose, when loading and reading cache */
                  *CacheContentPtr,
                  /* points into cache buffer when loading via record IO */
                  *CacheContentCurrentPtr,
                  /* heap memory in which the CGI variables are generated */
                  *CgiBufferPtr,
                  /* current position in CGI variable buffer */
                  *CgiBufferCurrentPtr,
                  /* script end-of-file string */
                  *CgiEofPtr,
                  /* points to dynamic buffer for body POSTed, PUT, etc. */
                  *ContentBufferPtr,
                  /* used as pointer into above buffer */
                  *ContentPtr,
                  /* buffer space for sys$output from DCL subprocess */
                  *DclSysOutputPtr,
                  /* pointer to other info when generating error report */
                  *ErrorOtherTextPtr,
                  /* pointer to error text/query-string */
                  *ErrorReportPtr,
                  /* pointer to explanation when generating error report */
                  *ErrorTextPtr,
                  /* "Accept:" header lines */
                  *HttpAcceptPtr,
                  /* "Accept-Charset:" header lines */
                  *HttpAcceptCharsetPtr,
                  /* "Accept-Language:" header lines */
                  *HttpAcceptLangPtr,
                  /* pointer to heap storage of "Authorization:" field */
                  *HttpAuthorizationPtr,
                  /* stream's MIME content type */
                  *HttpContentTypePtr,
                  /* "Cookie:" request header encountered */
                  *HttpCookiePtr,
                  /* "Forwarded:" header line */
                  *HttpForwardedPtr,
                  /* "Host:" header line */
                  *HttpHostPtr,
                  /* pointer to heap storage of "If-Modified-Since:" field */
                  *HttpIfModifiedSincePtr,
                  /* "Pragma:" request header encountered */
                  *HttpPragmaPtr,
                  /* pointer to heap storage of "Referer:" field */
                  *HttpRefererPtr,
                  /* pointer to heap storage of "User-Agent:" field */
                  *HttpUserAgentPtr,
                  /* pointer to heap storage of redirection location path */
                  *LocationPtr,
                  /* pointer to mapped path */
                  *MappedPathPtr,
                  /* pointer to heap storage client-read network buffering */
                  *NetReadBufferPtr,
                  /* pointer to heap storage of request path */
                  *PathInfoPtr,
                  /* pointer to heap storage of request query string */
                  *QueryStringPtr,
                  /* pointer to requested resource (path and query string) */
                  *ResourcePtr,
                  /* pointer to request header */
                  *RequestHeaderPtr,
                  /* pointer to response header */
                  *ResponseHeaderPtr,
                  /* buffer for data pointer */
                  *ResponseHeaderDataPtr,
                  /* rule mapping SET the path with a specific charset */
                  *SetPathCharsetPtr,
                  /* rule mapping SET the path with a specific content-type */
                  *SetPathContentTypePtr;

   unsigned char  /* GMT-adjusted string equivalent of 'BinaryTime' */
                  GmDateTime [48],
                  /* string reqpresenting the request's HTTP method */
                  HttpMethodName [METHOD_NAME_SIZE],
                  /* authenticated user name from "Authorization:" */
                  RemoteUser [MaxAuthUserNameLength+1],
                  /* storage when authenticating remote user */
                  RemoteUserPassword [MaxAuthPasswordLength+1],
                  /* parsed path storage */
                  RequestFileName [256],
                  /* */
                  RequestMappedFile [256],
                  /* */
                  RequestMappedScript [256],
                  /* mapped script name */
                  ScriptName [SCRIPT_NAME_SIZE];

   /* request is being used to fill a cache entry */
   struct CacheStruct  *CachePtr;

   /* AST function */
   void (*ResponseHeaderAstAddress)(struct RequestStruct*);

   /* storage for maintaining the thread's heap storage */
   struct HeapStruct  *HeapHeadPtr;
   struct HeapStruct  *HeapTailPtr;
   int  HeapSize;

   /* */
   struct ContentTypeStruct  ContentType;

   /* I/O status blocks */
   struct AnIOsb  NetReadIOsb,
                  NetWriteIOsb;

   /* RMS structures used for various file processing purposes */
   struct FAB  RequestParseFab;
   struct NAM  RequestParseNam;

   /**************/
   /* task lists */
   /**************/

   /* these two modules can have multiple, concurrent instances */
   struct  ListHeadStruct  DirTaskList,
                           SsiTaskList;

   /* these modules/tasks can only have one total or multiple serially */
   struct  DclTaskStruct      *DclTaskPtr;
   struct  DECnetTaskStruct   *DECnetTaskPtr;
   struct  DescrTaskStruct    *DescrTaskPtr;
   struct  DirTaskStruct      *DirTaskPtr;
   struct  FileTaskStruct     *FileTaskPtr;
   struct  GraphTaskStruct    *GraphTaskPtr;
   struct  HTAdminTaskStruct  *HTAdminTaskPtr;
   struct  IsMapTaskStruct    *IsMapTaskPtr;
   struct  MenuTaskStruct     *MenuTaskPtr;
   struct  PutTaskStruct      *PutTaskPtr;
   struct  SsiTaskStruct      *SsiTaskPtr;
   struct  UpdTaskStruct      *UpdTaskPtr;

   /****************************************/
   /* storage specific to output buffering */
   /****************************************/

   /* the head of the output buffer list */
   struct ListHeadStruct  OutputBufferList;

   /* pointer to heap storage structure for output buffering */
   struct ListEntryStruct  *OutputBufferStructPtr,
                           *OutputBufferNowStructPtr;

   /* pointer to any AST function */
   void  *OutputBufferAstFunctionPtr;

   boolean  /* all output buffers must be written to network */
            OutputBufferFlush;

   int  /* count of all data in all output buffers */
        OutputBufferCount,
        /* bytes still available in the current output buffer */
        OutputBufferRemaining;

   unsigned char  /* pointer to the start of the current buffer */
                  *OutputBufferPtr,
                  /* pointer used to point into 'OutputBuffer' storage */
                  *OutputBufferCurrentPtr;

   /****************************************************/
   /* storage specific to authentication/authorization */
   /****************************************************/

   boolean  /* user may change SYSUAF password */
            AuthCanChangeSysUafPwd,
            /* authenticated user may only access via https: (SSL) */
            AuthHttpsOnly,
            /* the authentication realm name is a synonym for the SYSUAF */
            AuthRealmIsVmsSynonym,
            /* the user was authenticated from the SYSUAF */
            AuthSysUafAuthenticated;

   unsigned short  /* length of the user profile pointed to */
                   AuthVmsUserProfileLength;

   unsigned long  /* value representation of scheme being used for challenge */
                  AuthChallengeScheme,
                  /* the capability flags of the authorization path */
                  AuthGroupCan,
                  /* the capability flags of the request (ACT ON THESE!) */
                  AuthRequestCan,
                  /* value representation of scheme being used for request */
                  AuthScheme,
                  /* the capability flags of the remote user */
                  AuthUserCan,
                  /* what everybody else can do on that path */
                  AuthWorldCan;

   unsigned char  /* BASIC challenge response header field */
                  *AuthBasicChallengePtr,
                  /* DIGEST challenge response header field */
                  *AuthDigestChallengePtr,
                  /* when digest authenticating, the client-supplied nonce */
                  *AuthDigestNoncePtr,
                  /* when digest authenticating, the client-supplied digest */
                  *AuthDigestResponsePtr,
                  /* when digest authenticating, the client-supplied URI */
                  *AuthDigestUriPtr,
                  /* pointer to authorization group name */
                  *AuthGroupPtr,
                  /* pointer to authorization realm name */
                  *AuthRealmPtr,
                  /* pointer to string of IP-address/username masks */
                  *AuthRestrictListPtr,
                  /* user profile is used by sys$check_access() */
                  *AuthVmsUserProfilePtr,
                  /* pointer to string of IP-address/username masks */
                  *AuthWorldRestrictListPtr;

   char  /* basic, digest, etc. */
         AuthType [16];


   /***************/
   /* zeroed end! */
   /***************/

   unsigned long  ZeroedEnd;
};

#ifdef __ALPHA
#   pragma member_alignment __restore
#endif

#endif /* WASD_H_LOADED */

/*****************************************************************************/
